@@ -155,9 +155,18 @@ tp_detect_wobbling(struct tp_dispatch *tp,
 	int dx, dy;
 	uint64_t dtime;
 
-	if (!(tp->queued & TOUCHPAD_EVENT_MOTION) || tp->hysteresis.enabled)
+	if (tp->nfingers_down != 1 ||
+	    tp->nfingers_down != tp->old_nfingers_down)
 		return;
 
+	if (tp->hysteresis.enabled)
+		return;
+
+	if (!(tp->queued & TOUCHPAD_EVENT_MOTION)) {
+		t->hysteresis.x_motion_history = 0;
+		return;
+	}
+
 	if (t->last_point.x == 0) { /* first invocation */
 		dx = 0;
 		dy = 0;
@@ -171,10 +180,7 @@ tp_detect_wobbling(struct tp_dispatch *tp,
 	tp->hysteresis.last_motion_time = time;
 	t->last_point = t->point;
 
-	if (dx == 0 && dy != 0) /* ignore y-only changes */
-		return;
-
-	if (dtime > ms2us(40)) {
+	if ((dx == 0 && dy != 0) || dtime > ms2us(40)) {
 		t->hysteresis.x_motion_history = 0;
 		return;
 	}
@@ -183,11 +189,13 @@ tp_detect_wobbling(struct tp_dispatch *tp,
 	if (dx > 0) { /* right move */
 		static const char r_l_r = 0x5; /* {Right, Left, Right} */
 
-
 		t->hysteresis.x_motion_history |= (1 << 2);
 		if (t->hysteresis.x_motion_history == r_l_r) {
 			tp->hysteresis.enabled = true;
-			evdev_log_debug(tp->device, "hysteresis enabled\n");
+			evdev_log_debug(tp->device,
+					"hysteresis enabled. "
+					"See %stouchpad_jitter.html for details\n",
+					HTTP_DOC_LINK);
 		}
 	}
 }
@@ -305,6 +313,7 @@ tp_new_touch(struct tp_dispatch *tp, struct tp_touch *t, uint64_t time)
 	t->dirty = true;
 	t->has_ended = false;
 	t->was_down = false;
+	t->palm.state = PALM_NONE;
 	t->state = TOUCH_HOVERING;
 	t->pinned.is_pinned = false;
 	t->time = time;
@@ -350,7 +359,8 @@ tp_maybe_end_touch(struct tp_dispatch *tp,
 		return;
 	case TOUCH_END:
 		evdev_log_bug_libinput(tp->device,
-				       "touch  already in TOUCH_END\n");
+				       "touch %d: already in TOUCH_END\n",
+				       t->index);
 		return;
 	case TOUCH_HOVERING:
 	case TOUCH_BEGIN:
@@ -391,7 +401,8 @@ tp_end_touch(struct tp_dispatch *tp, struct tp_touch *t, uint64_t time)
 {
 	if (t->state != TOUCH_MAYBE_END) {
 		evdev_log_bug_libinput(tp->device,
-				       "touch should be MAYBE_END, is %d\n",
+				       "touch %d should be MAYBE_END, is %d\n",
+				       t->index,
 				       t->state);
 		return;
 	}
@@ -600,7 +611,13 @@ tp_process_trackpoint_button(struct tp_dispatch *tp,
 {
 	struct evdev_dispatch *dispatch;
 	struct input_event event;
-	struct input_event syn_report = {{ 0, 0 }, EV_SYN, SYN_REPORT, 0 };
+	struct input_event syn_report = {
+		 .input_event_sec = 0,
+		 .input_event_usec = 0,
+		 .type = EV_SYN,
+		 .code = SYN_REPORT,
+		 .value = 0
+	};
 
 	if (!tp->buttons.trackpoint)
 		return;
@@ -608,7 +625,8 @@ tp_process_trackpoint_button(struct tp_dispatch *tp,
 	dispatch = tp->buttons.trackpoint->dispatch;
 
 	event = *e;
-	syn_report.time = e->time;
+	syn_report.input_event_sec = e->input_event_sec;
+	syn_report.input_event_usec = e->input_event_usec;
 
 	switch (event.code) {
 	case BTN_0:
@@ -743,7 +761,9 @@ tp_palm_tap_is_palm(const struct tp_dispatch *tp, const struct tp_touch *t)
 	if (!tp_palm_in_edge(tp, t))
 		return false;
 
-	evdev_log_debug(tp->device, "palm: palm-tap detected\n");
+	evdev_log_debug(tp->device,
+			"palm: touch %d: palm-tap detected\n",
+			t->index);
 	return true;
 }
 
@@ -771,7 +791,8 @@ tp_palm_detect_dwt_triggered(struct tp_dispatch *tp,
 		    t->palm.time > tp->dwt.keyboard_last_press_time) {
 			t->palm.state = PALM_NONE;
 			evdev_log_debug(tp->device,
-					"palm: touch released, timeout after typing\n");
+					"palm: touch %d released, timeout after typing\n",
+					t->index);
 		}
 	}
 
@@ -799,7 +820,7 @@ tp_palm_detect_trackpoint_triggered(struct tp_dispatch *tp,
 		    t->palm.time > tp->palm.trackpoint_last_event_time) {
 			t->palm.state = PALM_NONE;
 			evdev_log_debug(tp->device,
-				       "palm: touch released, timeout after trackpoint\n");
+				       "palm: touch %d released, timeout after trackpoint\n", t->index);
 		}
 	}
 
@@ -904,7 +925,8 @@ tp_palm_detect_touch_size_triggered(struct tp_dispatch *tp,
 	    t->minor > tp->palm.size_threshold) {
 		if (t->palm.state != PALM_TOUCH_SIZE)
 			evdev_log_debug(tp->device,
-					"palm: touch size exceeded\n");
+					"palm: touch %d size exceeded\n",
+					t->index);
 		t->palm.state = PALM_TOUCH_SIZE;
 		return true;
 	}
@@ -921,7 +943,8 @@ tp_palm_detect_edge(struct tp_dispatch *tp,
 		if (tp_palm_detect_multifinger(tp, t, time)) {
 			t->palm.state = PALM_NONE;
 			evdev_log_debug(tp->device,
-				  "palm: touch released, multiple fingers\n");
+				  "palm: touch %d released, multiple fingers\n",
+				  t->index);
 
 		/* If labelled a touch as palm, we unlabel as palm when
 		   we move out of the palm edge zone within the timeout, provided
@@ -930,7 +953,8 @@ tp_palm_detect_edge(struct tp_dispatch *tp,
 		} else if (tp_palm_detect_move_out_of_edge(tp, t, time)) {
 			t->palm.state = PALM_NONE;
 			evdev_log_debug(tp->device,
-				  "palm: touch released, out of edge zone\n");
+				  "palm: touch %d released, out of edge zone\n",
+				  t->index);
 		}
 		return false;
 	} else if (tp_palm_detect_multifinger(tp, t, time)) {
@@ -977,6 +1001,19 @@ tp_palm_detect_pressure_triggered(struct tp_dispatch *tp,
 	return t->palm.state == PALM_PRESSURE;
 }
 
+static bool
+tp_palm_detect_arbitration_triggered(struct tp_dispatch *tp,
+				     struct tp_touch *t,
+				     uint64_t time)
+{
+	if (!tp->arbitration.in_arbitration)
+		return false;
+
+	t->palm.state = PALM_ARBITRATION;
+
+	return true;
+}
+
 static void
 tp_palm_detect(struct tp_dispatch *tp, struct tp_touch *t, uint64_t time)
 {
@@ -986,6 +1023,9 @@ tp_palm_detect(struct tp_dispatch *tp, struct tp_touch *t, uint64_t time)
 	if (tp_palm_detect_pressure_triggered(tp, t, time))
 		goto out;
 
+	if (tp_palm_detect_arbitration_triggered(tp, t, time))
+		goto out;
+
 	if (tp_palm_detect_dwt_triggered(tp, t, time))
 		goto out;
 
@@ -1035,13 +1075,17 @@ out:
 	case PALM_TOUCH_SIZE:
 		palm_state = "touch size";
 		break;
+	case PALM_ARBITRATION:
+		palm_state = "arbitration";
+		break;
 	case PALM_NONE:
 	default:
 		abort();
 		break;
 	}
 	evdev_log_debug(tp->device,
-		  "palm: palm detected (%s)\n",
+		  "palm: touch %d, palm detected (%s)\n",
+		  t->index,
 		  palm_state);
 }
 
@@ -1115,7 +1159,8 @@ tp_thumb_detect(struct tp_dispatch *tp, struct tp_touch *t, uint64_t time)
 out:
 	if (t->thumb.state != state)
 		evdev_log_debug(tp->device,
-			  "thumb state: %s → %s\n",
+			  "thumb state: touch %d, %s → %s\n",
+			  t->index,
 			  thumb_state_to_str(state),
 			  thumb_state_to_str(t->thumb.state));
 }
@@ -1142,7 +1187,8 @@ tp_unhover_pressure(struct tp_dispatch *tp, uint64_t time)
 			if (t->state == TOUCH_HOVERING) {
 				if (t->pressure >= tp->pressure.high) {
 					evdev_log_debug(tp->device,
-							"pressure: begin touch\n");
+							"pressure: begin touch %d\n",
+							t->index);
 					/* avoid jumps when landing a finger */
 					tp_motion_history_reset(t);
 					tp_begin_touch(tp, t, time);
@@ -1156,7 +1202,8 @@ tp_unhover_pressure(struct tp_dispatch *tp, uint64_t time)
 				   tp->num_slots == 1) {
 				if (t->pressure < tp->pressure.low) {
 					evdev_log_debug(tp->device,
-							"pressure: end touch\n");
+							"pressure: end touch %d\n",
+							t->index);
 					tp_maybe_end_touch(tp, t, time);
 				}
 			}
@@ -1231,7 +1278,8 @@ tp_unhover_size(struct tp_dispatch *tp, uint64_t time)
 			if ((t->major > high && t->minor > low) ||
 			    (t->major > low && t->minor > high)) {
 				evdev_log_debug(tp->device,
-						"touch-size: begin touch\n");
+						"touch-size: begin touch %d\n",
+						t->index);
 				/* avoid jumps when landing a finger */
 				tp_motion_history_reset(t);
 				tp_begin_touch(tp, t, time);
@@ -1239,7 +1287,8 @@ tp_unhover_size(struct tp_dispatch *tp, uint64_t time)
 		} else {
 			if (t->major < low || t->minor < low) {
 				evdev_log_debug(tp->device,
-						"touch-size: end touch\n");
+						"touch-size: end touch %d\n",
+						t->index);
 				tp_maybe_end_touch(tp, t, time);
 			}
 		}
@@ -1457,7 +1506,8 @@ tp_detect_thumb_while_moving(struct tp_dispatch *tp)
 	/* Finger are too far apart or 2fg scrolling is disabled, mark
 	 * second finger as thumb */
 	evdev_log_debug(tp->device,
-			"touch is speed-based thumb\n");
+			"touch %d is speed-based thumb\n",
+			second->index);
 	second->thumb.state = THUMB_STATE_YES;
 }
 
@@ -1684,9 +1734,6 @@ tp_interface_process(struct evdev_dispatch *dispatch,
 {
 	struct tp_dispatch *tp = tp_dispatch(dispatch);
 
-	if (tp->ignore_events)
-		return;
-
 	switch (e->type) {
 	case EV_ABS:
 		if (tp->has_mt)
@@ -1709,7 +1756,7 @@ tp_interface_process(struct evdev_dispatch *dispatch,
 static void
 tp_remove_sendevents(struct tp_dispatch *tp)
 {
-	struct paired_keyboard *kbd;
+	struct evdev_paired_keyboard *kbd;
 
 	libinput_timer_cancel(&tp->palm.trackpoint_timer);
 	libinput_timer_cancel(&tp->dwt.keyboard_timer);
@@ -1719,9 +1766,8 @@ tp_remove_sendevents(struct tp_dispatch *tp)
 		libinput_device_remove_event_listener(
 					&tp->palm.trackpoint_listener);
 
-	ARRAY_FOR_EACH(tp->dwt.paired_keyboard, kbd) {
-		if (kbd->device)
-			libinput_device_remove_event_listener(&kbd->listener);
+	list_for_each(kbd, &tp->dwt.paired_keyboard_list, link) {
+		libinput_device_remove_event_listener(&kbd->listener);
 	}
 
 	if (tp->lid_switch.lid_switch)
@@ -1750,6 +1796,8 @@ tp_interface_destroy(struct evdev_dispatch *dispatch)
 {
 	struct tp_dispatch *tp = tp_dispatch(dispatch);
 
+	libinput_timer_cancel(&tp->arbitration.arbitration_timer);
+	libinput_timer_destroy(&tp->arbitration.arbitration_timer);
 	libinput_timer_destroy(&tp->palm.trackpoint_timer);
 	libinput_timer_destroy(&tp->dwt.keyboard_timer);
 	libinput_timer_destroy(&tp->tap.timer);
@@ -1793,8 +1841,16 @@ tp_clear_state(struct tp_dispatch *tp)
 }
 
 static void
-tp_suspend(struct tp_dispatch *tp, struct evdev_device *device)
+tp_suspend(struct tp_dispatch *tp,
+	   struct evdev_device *device,
+	   enum suspend_trigger trigger)
 {
+	if (tp->suspend_reason & trigger)
+		return;
+
+	if (tp->suspend_reason != 0)
+		goto out;
+
 	tp_clear_state(tp);
 
 	/* On devices with top softwarebuttons we don't actually suspend the
@@ -1808,6 +1864,9 @@ tp_suspend(struct tp_dispatch *tp, struct evdev_device *device)
 	} else {
 		evdev_device_suspend(device);
 	}
+
+out:
+	tp->suspend_reason |= trigger;
 }
 
 static void
@@ -1868,8 +1927,14 @@ tp_sync_slots(struct tp_dispatch *tp,
 }
 
 static void
-tp_resume(struct tp_dispatch *tp, struct evdev_device *device)
+tp_resume(struct tp_dispatch *tp,
+	  struct evdev_device *device,
+	  enum suspend_trigger trigger)
 {
+	tp->suspend_reason &= ~trigger;
+	if (tp->suspend_reason != 0)
+		return;
+
 	if (tp->buttons.has_topbuttons) {
 		/* tap state-machine is offline while suspended, reset state */
 		tp_clear_state(tp);
@@ -1883,31 +1948,6 @@ tp_resume(struct tp_dispatch *tp, struct evdev_device *device)
 	tp_sync_slots(tp, device);
 }
 
-#define NO_EXCLUDED_DEVICE NULL
-static void
-tp_resume_conditional(struct tp_dispatch *tp,
-		      struct evdev_device *device,
-		      struct evdev_device *excluded_device)
-{
-	if (tp->sendevents.current_mode == LIBINPUT_CONFIG_SEND_EVENTS_DISABLED)
-		return;
-
-	if (tp->sendevents.current_mode ==
-		    LIBINPUT_CONFIG_SEND_EVENTS_DISABLED_ON_EXTERNAL_MOUSE) {
-		struct libinput_device *dev;
-
-		list_for_each(dev, &device->base.seat->devices_list, link) {
-			struct evdev_device *d = evdev_device(dev);
-			if (d != excluded_device &&
-			    (d->tags & EVDEV_TAG_EXTERNAL_MOUSE)) {
-				return;
-			}
-		}
-	}
-
-	tp_resume(tp, device);
-}
-
 static void
 tp_trackpoint_timeout(uint64_t now, void *data)
 {
@@ -2094,8 +2134,8 @@ tp_dwt_pair_keyboard(struct evdev_device *touchpad,
 		     struct evdev_device *keyboard)
 {
 	struct tp_dispatch *tp = (struct tp_dispatch*)touchpad->dispatch;
-	struct paired_keyboard *kbd;
-	bool found = false;
+	struct evdev_paired_keyboard *kbd;
+	size_t count = 0;
 
 	if ((keyboard->tags & EVDEV_TAG_KEYBOARD) == 0)
 		return;
@@ -2103,25 +2143,25 @@ tp_dwt_pair_keyboard(struct evdev_device *touchpad,
 	if (!tp_want_dwt(touchpad, keyboard))
 		return;
 
-	ARRAY_FOR_EACH(tp->dwt.paired_keyboard, kbd) {
-		if (kbd->device)
-			continue;
-
-		found = true;
-		libinput_device_add_event_listener(&keyboard->base,
-						   &kbd->listener,
-						   tp_keyboard_event, tp);
-		kbd->device = keyboard;
-		evdev_log_debug(touchpad,
-				"palm: dwt activated with %s<->%s\n",
-				touchpad->devname,
-				keyboard->devname);
-		break;
+	list_for_each(kbd, &tp->dwt.paired_keyboard_list, link) {
+		count++;
+		if (count > 3) {
+			evdev_log_info(touchpad,
+				       "too many internal keyboards for dwt\n");
+			break;
+		}
 	}
 
-	if (!found)
-		evdev_log_bug_libinput(touchpad,
-				       "too many internal keyboards for dwt\n");
+	kbd = zalloc(sizeof(*kbd));
+	kbd->device = keyboard;
+	libinput_device_add_event_listener(&keyboard->base,
+					   &kbd->listener,
+					   tp_keyboard_event, tp);
+	list_insert(&tp->dwt.paired_keyboard_list, &kbd->link);
+	evdev_log_debug(touchpad,
+			"palm: dwt activated with %s<->%s\n",
+			touchpad->devname,
+			keyboard->devname);
 }
 
 static void
@@ -2152,34 +2192,54 @@ tp_pair_trackpoint(struct evdev_device *touchpad,
 }
 
 static void
-tp_switch_event(uint64_t time, struct libinput_event *event, void *data)
+tp_lid_switch_event(uint64_t time, struct libinput_event *event, void *data)
 {
 	struct tp_dispatch *tp = data;
 	struct libinput_event_switch *swev;
-	const char *which = NULL;
 
 	if (libinput_event_get_type(event) != LIBINPUT_EVENT_SWITCH_TOGGLE)
 		return;
 
 	swev = libinput_event_get_switch_event(event);
-
-	switch (libinput_event_switch_get_switch(swev)) {
-	case LIBINPUT_SWITCH_LID:
-		which = "lid";
-		break;
-	case LIBINPUT_SWITCH_TABLET_MODE:
-		which = "tablet-mode";
-		break;
-	}
+	if (libinput_event_switch_get_switch(swev) != LIBINPUT_SWITCH_LID)
+		return;
 
 	switch (libinput_event_switch_get_switch_state(swev)) {
 	case LIBINPUT_SWITCH_STATE_OFF:
-		tp_resume_conditional(tp, tp->device, NO_EXCLUDED_DEVICE);
-		evdev_log_debug(tp->device, "%s: resume touchpad\n", which);
+		tp_resume(tp, tp->device, SUSPEND_LID);
+		evdev_log_debug(tp->device, "lid: resume touchpad\n");
 		break;
 	case LIBINPUT_SWITCH_STATE_ON:
-		tp_suspend(tp, tp->device);
-		evdev_log_debug(tp->device, "%s: suspending touchpad\n", which);
+		tp_suspend(tp, tp->device, SUSPEND_LID);
+		evdev_log_debug(tp->device, "lid: suspending touchpad\n");
+		break;
+	}
+}
+
+static void
+tp_tablet_mode_switch_event(uint64_t time,
+			    struct libinput_event *event,
+			    void *data)
+{
+	struct tp_dispatch *tp = data;
+	struct libinput_event_switch *swev;
+
+	if (libinput_event_get_type(event) != LIBINPUT_EVENT_SWITCH_TOGGLE)
+		return;
+
+	swev = libinput_event_get_switch_event(event);
+	if (libinput_event_switch_get_switch(swev) !=
+	    LIBINPUT_SWITCH_TABLET_MODE)
+		return;
+
+	switch (libinput_event_switch_get_switch_state(swev)) {
+	case LIBINPUT_SWITCH_STATE_OFF:
+		tp_resume(tp, tp->device, SUSPEND_TABLET_MODE);
+		evdev_log_debug(tp->device, "tablet-mode: resume touchpad\n");
+		break;
+	case LIBINPUT_SWITCH_STATE_ON:
+		tp_suspend(tp, tp->device, SUSPEND_TABLET_MODE);
+		evdev_log_debug(tp->device, "tablet-mode: suspending touchpad\n");
 		break;
 	}
 }
@@ -2201,7 +2261,7 @@ tp_pair_lid_switch(struct evdev_device *touchpad,
 
 		libinput_device_add_event_listener(&lid_switch->base,
 						   &tp->lid_switch.listener,
-						   tp_switch_event, tp);
+						   tp_lid_switch_event, tp);
 		tp->lid_switch.lid_switch = lid_switch;
 	}
 }
@@ -2225,13 +2285,13 @@ tp_pair_tablet_mode_switch(struct evdev_device *touchpad,
 
 	libinput_device_add_event_listener(&tablet_mode_switch->base,
 				&tp->tablet_mode_switch.listener,
-				tp_switch_event, tp);
+				tp_tablet_mode_switch_event, tp);
 	tp->tablet_mode_switch.tablet_mode_switch = tablet_mode_switch;
 
 	if (evdev_device_switch_get_state(tablet_mode_switch,
 					  LIBINPUT_SWITCH_TABLET_MODE)
 		    == LIBINPUT_SWITCH_STATE_ON) {
-		tp_suspend(tp, touchpad);
+		tp_suspend(tp, touchpad, SUSPEND_TABLET_MODE);
 	}
 }
 
@@ -2251,7 +2311,7 @@ tp_interface_device_added(struct evdev_device *device,
 		return;
 
 	if (added_device->tags & EVDEV_TAG_EXTERNAL_MOUSE)
-		tp_suspend(tp, device);
+		tp_suspend(tp, device, SUSPEND_EXTERNAL_MOUSE);
 }
 
 static void
@@ -2259,7 +2319,7 @@ tp_interface_device_removed(struct evdev_device *device,
 			    struct evdev_device *removed_device)
 {
 	struct tp_dispatch *tp = (struct tp_dispatch*)device->dispatch;
-	struct paired_keyboard *kbd;
+	struct evdev_paired_keyboard *kbd, *tmp;
 
 	if (removed_device == tp->buttons.trackpoint) {
 		/* Clear any pending releases for the trackpoint */
@@ -2273,10 +2333,9 @@ tp_interface_device_removed(struct evdev_device *device,
 		tp->buttons.trackpoint = NULL;
 	}
 
-	ARRAY_FOR_EACH(tp->dwt.paired_keyboard, kbd) {
+	list_for_each_safe(kbd, tmp, &tp->dwt.paired_keyboard_list, link) {
 		if (kbd->device == removed_device) {
-			libinput_device_remove_event_listener(&kbd->listener);
-			kbd->device = NULL;
+			evdev_paired_keyboard_destroy(kbd);
 			tp->dwt.keyboard_active = false;
 		}
 	}
@@ -2285,17 +2344,32 @@ tp_interface_device_removed(struct evdev_device *device,
 		libinput_device_remove_event_listener(
 					&tp->lid_switch.listener);
 		tp->lid_switch.lid_switch = NULL;
+		tp_resume(tp, device, SUSPEND_LID);
 	}
 
 	if (removed_device == tp->tablet_mode_switch.tablet_mode_switch) {
 		libinput_device_remove_event_listener(
 					&tp->tablet_mode_switch.listener);
 		tp->tablet_mode_switch.tablet_mode_switch = NULL;
+		tp_resume(tp, device, SUSPEND_TABLET_MODE);
 	}
 
-	/* removed_device is still in the device list at this point, so we
-	 * need to exclude it from the tp_resume_conditional */
-	tp_resume_conditional(tp, device, removed_device);
+	if (tp->sendevents.current_mode ==
+		    LIBINPUT_CONFIG_SEND_EVENTS_DISABLED_ON_EXTERNAL_MOUSE) {
+		struct libinput_device *dev;
+		bool found = false;
+
+		list_for_each(dev, &device->base.seat->devices_list, link) {
+			struct evdev_device *d = evdev_device(dev);
+			if (d != removed_device &&
+			    (d->tags & EVDEV_TAG_EXTERNAL_MOUSE)) {
+				found = true;
+				break;
+			}
+		}
+		if (!found)
+			tp_resume(tp, device, SUSPEND_EXTERNAL_MOUSE);
+	}
 }
 
 static inline void
@@ -2375,21 +2449,42 @@ evdev_tag_touchpad(struct evdev_device *device,
 	}
 }
 
+static void
+tp_arbitration_timeout(uint64_t now, void *data)
+{
+	struct tp_dispatch *tp = data;
+
+	if (tp->arbitration.in_arbitration)
+		tp->arbitration.in_arbitration = false;
+}
+
 static void
 tp_interface_toggle_touch(struct evdev_dispatch *dispatch,
 			  struct evdev_device *device,
-			  bool enable)
+			  bool enable,
+			  uint64_t time)
 {
 	struct tp_dispatch *tp = tp_dispatch(dispatch);
-	bool ignore_events = !enable;
+	bool arbitrate = !enable;
 
-	if (ignore_events == tp->ignore_events)
+	if (arbitrate == tp->arbitration.in_arbitration)
 		return;
 
-	if (ignore_events)
+	if (arbitrate) {
+		libinput_timer_cancel(&tp->arbitration.arbitration_timer);
 		tp_clear_state(tp);
-
-	tp->ignore_events = ignore_events;
+		tp->arbitration.in_arbitration = true;
+	} else {
+		/* if in-kernel arbitration is in use and there is a touch
+		 * and a pen in proximity, lifting the pen out of proximity
+		 * causes a touch being for the touch. On a hand-lift the
+		 * proximity out precedes the touch up by a few ms, so we
+		 * get what looks like a tap. Fix this by delaying
+		 * arbitration by just a little bit so that any touch in
+		 * event is caught as palm touch. */
+		libinput_timer_set(&tp->arbitration.arbitration_timer,
+				   time + ms2us(90));
+	}
 }
 
 static struct evdev_dispatch_interface tp_interface = {
@@ -2408,10 +2503,12 @@ static struct evdev_dispatch_interface tp_interface = {
 
 static void
 tp_init_touch(struct tp_dispatch *tp,
-	      struct tp_touch *t)
+	      struct tp_touch *t,
+	      unsigned int index)
 {
 	t->tp = tp;
 	t->has_ended = true;
+	t->index = index;
 }
 
 static inline void
@@ -2495,7 +2592,7 @@ tp_init_slots(struct tp_dispatch *tp,
 	tp->touches = zalloc(tp->ntouches * sizeof(struct tp_touch));
 
 	for (i = 0; i < tp->ntouches; i++)
-		tp_init_touch(tp, &tp->touches[i]);
+		tp_init_touch(tp, &tp->touches[i], i);
 
 	tp_sync_slots(tp, device);
 
@@ -2750,6 +2847,12 @@ tp_is_tpkb_combo_below(struct evdev_device *device)
 		layout == TPKBCOMBO_LAYOUT_BELOW;
 }
 
+static inline bool
+tp_is_tablet(struct evdev_device *device)
+{
+	return device->tags & EVDEV_TAG_TABLET_TOUCHPAD;
+}
+
 static void
 tp_init_dwt(struct tp_dispatch *tp,
 	    struct evdev_device *device)
@@ -2847,9 +2950,6 @@ tp_init_palmdetect_size(struct tp_dispatch *tp,
 	const char *prop;
 	int threshold;
 
-	if (!tp->touch_size.use_touch_size)
-		return;
-
 	prop = udev_device_get_property_value(device->udev_device,
 					      "LIBINPUT_ATTR_PALM_SIZE_THRESHOLD");
 	if (!prop)
@@ -2867,6 +2967,23 @@ tp_init_palmdetect_size(struct tp_dispatch *tp,
 	tp->palm.size_threshold = threshold;
 }
 
+static inline void
+tp_init_palmdetect_arbitration(struct tp_dispatch *tp,
+			       struct evdev_device *device)
+{
+	char timer_name[64];
+
+	snprintf(timer_name,
+		 sizeof(timer_name),
+		  "%s arbitration",
+		  evdev_device_get_sysname(device));
+	libinput_timer_init(&tp->arbitration.arbitration_timer,
+			    tp_libinput_context(tp),
+			    timer_name,
+			    tp_arbitration_timeout, tp);
+	tp->arbitration.in_arbitration = false;
+}
+
 static void
 tp_init_palmdetect(struct tp_dispatch *tp,
 		   struct evdev_device *device)
@@ -2876,18 +2993,23 @@ tp_init_palmdetect(struct tp_dispatch *tp,
 	tp->palm.left_edge = INT_MIN;
 	tp->palm.upper_edge = INT_MIN;
 
+	tp_init_palmdetect_arbitration(tp, device);
+
 	if (device->tags & EVDEV_TAG_EXTERNAL_TOUCHPAD &&
-	    !tp_is_tpkb_combo_below(device))
+	    !tp_is_tpkb_combo_below(device) &&
+	    !tp_is_tablet(device))
 		return;
 
-	tp->palm.monitor_trackpoint = true;
+	if (!tp_is_tablet(device))
+		tp->palm.monitor_trackpoint = true;
 
 	if (libevdev_has_event_code(device->evdev,
 				    EV_ABS,
 				    ABS_MT_TOOL_TYPE))
 		tp->palm.use_mt_tool = true;
 
-	tp_init_palmdetect_edge(tp, device);
+	if (!tp_is_tablet(device))
+		tp_init_palmdetect_edge(tp, device);
 	tp_init_palmdetect_pressure(tp, device);
 	tp_init_palmdetect_size(tp, device);
 }
@@ -2917,6 +3039,25 @@ tp_init_sendevents(struct tp_dispatch *tp,
 			    tp_keyboard_timeout, tp);
 }
 
+static int
+tp_read_thumb_pressure_prop(struct tp_dispatch *tp,
+			    const struct evdev_device *device)
+{
+	struct udev_device *udev_device = device->udev_device;
+	const char *prop;
+	int threshold;
+	const int default_thumb_threshold = 0;
+
+	prop = udev_device_get_property_value(udev_device,
+			      "LIBINPUT_ATTR_THUMB_PRESSURE_THRESHOLD");
+	if (!prop)
+		return default_thumb_threshold;
+
+	threshold = parse_thumb_pressure_property(prop);
+
+	return threshold > 0 ? threshold : default_thumb_threshold;
+}
+
 static void
 tp_init_thumb(struct tp_dispatch *tp)
 {
@@ -2925,8 +3066,7 @@ tp_init_thumb(struct tp_dispatch *tp)
 	double w = 0.0, h = 0.0;
 	struct device_coords edges;
 	struct phys_coords mm = { 0.0, 0.0 };
-	int xres, yres;
-	double threshold;
+	int threshold;
 
 	if (!tp->buttons.is_clickpad)
 		return;
@@ -2955,20 +3095,11 @@ tp_init_thumb(struct tp_dispatch *tp)
 	if (!abs)
 		goto out;
 
-	if (abs->maximum - abs->minimum < 255)
+	threshold = tp_read_thumb_pressure_prop(tp, device);
+	if (threshold == 0)
 		goto out;
 
-	/* Our reference touchpad is the T440s with 42x42 resolution.
-	 * Higher-res touchpads exhibit higher pressure for the same
-	 * interaction. On the T440s, the threshold value is 100, you don't
-	 * reach that with a normal finger interaction.
-	 * Note: "thumb" means massive touch that should not interact, not
-	 * "using the tip of my thumb for a pinch gestures".
-	 */
-	xres = tp->device->abs.absinfo_x->resolution;
-	yres = tp->device->abs.absinfo_y->resolution;
-	threshold = 100.0 * hypot(xres, yres)/hypot(42, 42);
-	tp->thumb.threshold = max(100, threshold);
+	tp->thumb.threshold = threshold;
 
 out:
 	evdev_log_debug(device,
@@ -3052,7 +3183,12 @@ tp_init_hysteresis(struct tp_dispatch *tp)
 
 	tp->hysteresis.margin.x = xmargin;
 	tp->hysteresis.margin.y = ymargin;
-	tp->hysteresis.enabled = false;
+	tp->hysteresis.enabled = (ax->fuzz || ay->fuzz);
+	if (tp->hysteresis.enabled)
+		evdev_log_debug(tp->device,
+				"hysteresis enabled. "
+				"See %stouchpad_jitter.html for details\n",
+				HTTP_DOC_LINK);
 }
 
 static void
@@ -3060,13 +3196,11 @@ tp_init_pressure(struct tp_dispatch *tp,
 		 struct evdev_device *device)
 {
 	const struct input_absinfo *abs;
-	unsigned int code = ABS_PRESSURE;
+	unsigned int code;
 	const char *prop;
 	int hi, lo;
 
-	if (tp->has_mt)
-		code = ABS_MT_PRESSURE;
-
+	code = tp->has_mt ? ABS_MT_PRESSURE : ABS_PRESSURE;
 	if (!libevdev_has_event_code(device->evdev, EV_ABS, code)) {
 		tp->pressure.use_pressure = false;
 		return;
@@ -3129,17 +3263,17 @@ tp_init_touch_size(struct tp_dispatch *tp,
 		return false;
 	}
 
+	prop = udev_device_get_property_value(device->udev_device,
+					      "LIBINPUT_ATTR_TOUCH_SIZE_RANGE");
+	if (!prop)
+		return false;
+
 	if (libevdev_get_num_slots(device->evdev) < 5) {
 		evdev_log_bug_libinput(device,
 			       "Expected 5+ slots for touch size detection\n");
 		return false;
 	}
 
-	prop = udev_device_get_property_value(device->udev_device,
-					      "LIBINPUT_ATTR_TOUCH_SIZE_RANGE");
-	if (!prop)
-		return false;
-
 	if (!parse_range_property(prop, &hi, &lo)) {
 		evdev_log_bug_client(device,
 				     "discarding invalid touch size range '%s'\n",
@@ -3158,7 +3292,9 @@ tp_init_touch_size(struct tp_dispatch *tp,
 	tp->touch_size.high = hi;
 	tp->touch_size.use_touch_size = true;
 
-	evdev_log_debug(device, "using size-based touch detection\n");
+	evdev_log_debug(device,
+			"using size-based touch detection (%d:%d)\n",
+			hi, lo);
 
 	return true;
 }
@@ -3172,6 +3308,7 @@ tp_init(struct tp_dispatch *tp,
 	tp->base.dispatch_type = DISPATCH_TOUCHPAD;
 	tp->base.interface = &tp_interface;
 	tp->device = device;
+	list_init(&tp->dwt.paired_keyboard_list);
 
 	if (!tp_pass_sanity_check(tp, device))
 		return false;
@@ -3182,9 +3319,7 @@ tp_init(struct tp_dispatch *tp,
 		return false;
 
 	evdev_device_init_abs_range_warnings(device);
-
-	if (device->model_flags & EVDEV_MODEL_APPLE_TOUCHPAD)
-		use_touch_size = tp_init_touch_size(tp, device);
+	use_touch_size = tp_init_touch_size(tp, device);
 
 	if (!use_touch_size)
 		tp_init_pressure(tp, device);
@@ -3235,8 +3370,8 @@ tp_suspend_conditional(struct tp_dispatch *tp,
 	list_for_each(dev, &device->base.seat->devices_list, link) {
 		struct evdev_device *d = evdev_device(dev);
 		if (d->tags & EVDEV_TAG_EXTERNAL_MOUSE) {
-			tp_suspend(tp, device);
-			return;
+			tp_suspend(tp, device, SUSPEND_EXTERNAL_MOUSE);
+			break;
 		}
 	}
 }
@@ -3258,13 +3393,16 @@ tp_sendevents_set_mode(struct libinput_device *device,
 
 	switch(mode) {
 	case LIBINPUT_CONFIG_SEND_EVENTS_ENABLED:
-		tp_resume(tp, evdev);
+		tp_resume(tp, evdev, SUSPEND_SENDEVENTS);
+		tp_resume(tp, evdev, SUSPEND_EXTERNAL_MOUSE);
 		break;
 	case LIBINPUT_CONFIG_SEND_EVENTS_DISABLED:
-		tp_suspend(tp, evdev);
+		tp_suspend(tp, evdev, SUSPEND_SENDEVENTS);
+		tp_resume(tp, evdev, SUSPEND_EXTERNAL_MOUSE);
 		break;
 	case LIBINPUT_CONFIG_SEND_EVENTS_DISABLED_ON_EXTERNAL_MOUSE:
 		tp_suspend_conditional(tp, evdev);
+		tp_resume(tp, evdev, SUSPEND_SENDEVENTS);
 		break;
 	default:
 		return LIBINPUT_CONFIG_STATUS_UNSUPPORTED;
